// usb.h -- interface to USBUHCI[L] DOS USB drivers by Bret Johnson


#ifndef USB_H
#define USB_H


///////////////
// Constants //
///////////////


// Int 14 Request Constants
// They go in the various Int 14h Request Structure Parameters,
//   or are sent as return values (error codes & such).

//--------------------------------------------------------------------------
//Int 14h Request Types
//--------------------------------------------------------------------------
#define I14RRTHostClass         0x00 //Host/System/OS Class
#define   I14RRTGetHostSWInfo   0x01 //Get Host Software Info
#define   I14RRTGetHostHWInfo   0x02 //Get Host Hardware Info
#define   I14RRTGetHostVendInfo 0x03 //Get Host Vendor Info
#define   I14RRTGetHostStatus   0x04 //Get Current Host Status
#define   I14RRTHostRun         0x08 //Start/Run/Resume Host
#define   I14RRTHostStop        0x09 //Stop Host
#define   I14RRTHostReset       0x0A //Reset Host
#define   I14RRTHostSuspend     0x0B //Global Suspend Host
#define   I14RRTHostResume      0x0C //Force Global Resume on Host
#define I14RRTTimingClass       0x10 //Frame Timing Class
#define   I14RRTRegTmgOwner     0x11 //Register as Timing Owner
#define   I14RRTUnRegTmgOwner   0x12 //UnRegister as Timing Owner
#define   I14RRTIncTiming       0x13 //Increment (Slow Down) Frame Timing
#define   I14RRTDecTiming       0x14 //Decrement (Speed Up) Frame Timing
#define   I14RRTChangeTiming    0x15 //Change Frame Timing (by Large Amount)
#define I14RRTHubClass          0x20 //Hub Class
#define   I14RRTGetDvcHubInfo   0x21 //Get Hub Info for Device
#define   I14RRTNewDvcConn      0x24 //Hub has Detected new Device
#define   I14RRTDvcDisc         0x25 //Device has been Disconnected
#define   I14RRTSendHubChar     0x27 //Send Hub Characteristics to Host
#define   I14RRTEnableHubPort   0x28 //Enable Device given Hub & Port
#define   I14RRTDisableHubPort  0x29 //Disable Device given Hub & Port
#define   I14RRTResetHubPort    0x2A //Reset Device given Hub & Port
#define   I14RRTSuspendHubPort  0x2B //Suspend Device given Hub & Port
#define   I14RRTResumeHubPort   0x2C //Resume Device given Hub & Port
#define   I14RRTPwrOnHubPort    0x2D //Power On Device given Hub & Port
#define   I14RRTPwrOffHubPort   0x2E //Power Off Device given Hub & Port
#define   I14RRTPwrResetHubPort 0x2F //Power Reset Device given Hub & Port
#define I14RRTTPowerClass       0x30 //Power Class
#define   I14RRTGetDvcPowerInfo 0x31 //Get Power Info for Device
#define   I14RRTPwrOnDevice     0x3D //Power On Device given Dvc Addr
#define   I14RRTPwrOffDevice    0x3E //Power Off Device given Dvc Addr
#define   I14RRTPwrResetDevice  0x3F //Power Reset Device given Dvc Addr
#define I14RRTDeviceClass       0x40 //Device Class
#define   I14RRTGetDvcClassInfo 0x41 //Get Device Class Info
#define   I14RRTGetDvcVendInfo  0x42 //Get Device Vendor Info
#define   I14RRTGetDvcStatus    0x43 //Get Device Status
#define   I14RRTEnableDevice    0x48 //Enable/Resume Device given Dvc Addr
#define   I14RRTDisableDevice   0x49 //Disable Device given Dvc Addr
#define   I14RRTResetDevice     0x4A //Reset Device given Dvc Addr
#define   I14RRTSuspendDevice   0x4B //Suspend Device given Dvc Addr
#define   I14RRTResumeDevice    0x4C //Resume Device given Dvc Addr
#define I14RRTConfigClass       0x50 //Configuration Class
#define   I14RRTSetNewConfig    0x58 //Set/Change Device Config Value
#define I14RRTInterfaceClass    0x60 //Interface Class
#define   I14RRTFindRegIntf     0x62 //Look for Registered Interface
#define   I14RRTFindUnRegIntf   0x63 //Look for Unregistered Interface
#define   I14RRTRegIntfOwner    0x64 //Register as Interface Owner
#define   I14RRTUnRegIntfOwner  0x65 //Unregister as Interface Owner
#define   I14RRTIntfDontLook    0x68 //Existing Interface Owner Don't Look
#define I14RRTAltIntfClass      0x70 //Alternate Interface Class
#define   I14RRTGetAltIntfInfo  0x71 //Get Alternate Interface Info
#define I14RRTEndPointClass     0x80 //End Point Class
#define I14RRTPacketClass       0x90 //Packet Class
#define   I14RRTDoIsoch         0x94 //Schedule Isochronous Transaction
#define   I14RRTDoInterruptPer  0x95 //Schedule Periodic Interrupt
#define   I14RRTDoControl       0x96 //Schedule Control/Setup Request
#define   I14RRTDoBulk          0x97 //Schedule Bulk Transaction
#define   I14RRTDoInterrupt1T   0x98 //Schedule One-Time Interrupt
#define   I14RRTCloseHandle     0x9C //Close/Remove Scheduled Transaction
#define   I14RRTChangeIntPer    0x9D //Change Periodicity of Interrupt
#define   I14RRTGetTransStatus  0x9F //Get Status of Packet Transaction
#define I14RRTMiscClass         0xA0 //Miscellaneous Class
#define   I14RRTLargeCallDone   0xA1 //Large (Complicated) Call Complete
#define   I14RRTBeepSpeaker     0xAF //Beep the Speaker
#define I14RRTInterHostClass    0xE0 //Inter-Host Communication Class
#define   I14RRTCopyNDTable     0xE5 //Copy New Device Ownership Table
#define I14RRTInternalClass     0xF0 //Host Internal/Troubleshooting Class

//--------------------------------------------------------------------------
//Int 14h Flags
//--------------------------------------------------------------------------
#define I14RFlagIn            0x01 //In Direction
#define I14RFlagLowSpeed      0x02 //Low-Speed Device
#define I14RFlagHiSpeed       0x04 //High-Speed Device
#define I14RFlagNoRetries     0x10 //No Auto Retries for Control
#define I14RFlagShortPktOK    0x20 //No Retries for Short Packets
#define I14RFlagSpecificFrame 0x40 //Use Specific Frame Number
#define I14RFlagAddrIsPhys    0x80 //Data Address is Physical

//--------------------------------------------------------------------------
//Int 14h Error Codes
//--------------------------------------------------------------------------
#define I14RErrSuccess        0x0000 //No Error at all (Success!)
#define I14RErrResources      0x0001 //Insufficient Resources/Memory
#define I14RErrHostHardware   0x0002 //Unknown Hardware Problem with Host
#define I14RErrReqTypeNoSupt  0x0003 //Request Type Not Supported
#define I14RErrRequestType    0x0011 //Invalid Request Type
#define I14RErrFlags          0x0012 //Invalid Flags
#define I14RErrHostIndex      0x0013 //Invalid Host Index
#define I14RErrDeviceAddress  0x0014 //Invalid Device Address
#define I14RErrHubAddress     0x0014 //Invalid Hub Address
#define I14RErrEndPoint       0x0015 //Invalid End Point
#define I14RErrHubPort        0x0015 //Invalid Hub Port
#define I14RErrAltIntf        0x0015 //Invalid Alternate Interface
#define I14RErrConfigValue    0x0016 //Invalid Configuration Value
#define I14RErrCloseID        0x0016 //Invalid Closure ID
#define I14RErrInterfaceNum   0x0017 //Invalid Interface Number
#define I14RErrSearchIndex    0x0018 //Invalid Search Index
#define I14RErrVendorProdID   0x0019 //Invalid Vendor/Product ID
#define I14RErrDvcIntf        0x001A //Invalid Device/Interface Value
#define I14RErrRequestHandle  0x001B //Invalid Request Handle
#define I14RErrPeriodicity    0x001C //Invalid Periodicity
#define I14RErrBeepFrequency  0x001C //Invalid Beep Frequency
#define I14RErrTimeOut        0x001D //Invalid Timeout
#define I14RErrDataAddress    0x001E //Invalid Data Address
#define I14RErrDataSize       0x001F //Invalid Data Size
#define I14RErrCallBackAddr   0x0021 //Invalid Call Back Address
#define I14RErrLCRtnCode      0x0021 //Invalid Large Call Return Code
#define I14RErrUserPktID      0x0022 //Invalid User Packet ID
#define I14RErrSetupReqData   0x0023 //Invalid Setup Request Data
#define I14RErrFrameTiming    0x0024 //Invalid Frame Timing
#define I14RErrFrameIndex     0x0024 //Invalid Frame Index
#define I14RErrIsochArray     0x0025 //Invalid Entry in Isoch Array Data
#define I14RErrHostHalted     0x0031 //Host is Halted/Suspended
#define I14RErrIntfNotFound   0x0032 //Interface Not Found
#define I14RErrIntfInConfig   0x0033 //Interface Already Being Configured
#define I14RErrIntfAlreadyReg 0x0034 //Interface Already Registered
#define I14RErrHostCallInUse  0x0035 //Host Run/Stop/etc. Already Running
#define I14RErrNoPortPower    0x0036 //Per-Port-Power not Supported by Hub
#define I14RErrAlreadyBeeping 0x0037 //Beep Request Already Running
#define I14RErrDebugMode      0x0041 //Tried Single-Step with no Debug Mode?
#define I14RErrTimingOwned    0x0051 //Problem with Frame Timing Ownership
#define I14RErrTimingLimit    0x0052 //Frame Timing Already at a Limit
#define I14RErrTmgChgInProg   0x0053 //Timing Change Already in Progress

//--------------------------------------------------------------------------
//Error Codes Returned on Completion of Transfer Descriptor (AX)
//--------------------------------------------------------------------------
#define TDStsOK             0x0000 //ACK Received (TD completed OK / no errors)
#define TDStsNAKReceived    0x0001 //NAK Received
#define TDStsStalled        0x0002 //TD is Stalled
#define TDStsTimeout        0x0004 //TD has timed out (Bulk/Control)
#define TDStsOverDue        0x0008 //TD is OverDue (Int/Isoch)
#define TDStsShortPacket    0x0010 //TD Short Packet Detected (w/ SPD Enabled)
#define TDStsBabbleDetected 0x0020 //Babble Detected
#define TDStsCRCTOReceived  0x0040 //CRC/TimeOut Error Received
#define TDStsBitStuffError  0x0080 //Rx Data contained > 6 ones in a row
#define TDStsDataBuffErr    0x0100 //Data Buffer Error
#define TDStsControlSetup   0x1000 //Error occurred during Control Setup Stage
#define TDStsLargeCallErr   0x8000 //Error during Large Call

//--------------------------------------------------------------------------
//Codes Returned by Int 14h Get Host Hardware Info Request (BL)
//--------------------------------------------------------------------------
  //Low Nibble (not bit-mapped)
#define USBHostTypeUHCI 0x01 //Universal Host Controller Interface
#define USBHostTypeOHCI 0x02 //Open Host Controller Interface
#define USBHostTypeEHCI 0x03 //Enhanced Host Controller Interface (USB 2.0)

  //High Nibble (not bit-mapped)
#define USBBusTypePCI   0x10 //Peripheral Component Interconnect

//--------------------------------------------------------------------------
//Codes Returned by Int 14h Get Device Info Request (DL)
//--------------------------------------------------------------------------
#define LowSpeedFlag 0x01 //Device is Low Speed

//--------------------------------------------------------------------------
//Codes Returned by Int 14h Get Device Status Request (CH)
//--------------------------------------------------------------------------
#define BadDevice        0x01 //Bad Device
#define RWakeupSupported 0x02 //Remote Wakeup Supported
#define RWakeupEnabled   0x04 //Remote Wakeup Enabled
#define TestModeEnabled  0x08 //Test Mode Enabled
#define BeingConfiged    0x80 //Device is currently being Configured

//--------------------------------------------------------------------------
//Codes Returned by Int 14h Get Alternate Interface Info Request (CL)
//--------------------------------------------------------------------------
#define AltIntfSelected 0x01 //Alternate Interface is Selected
#define AltIntfOwned    0x02 //(Alternate) Interface is Owned

//--------------------------------------------------------------------------
//Codes Returned by Int 14h Get Device Power Info Request (CL)
//--------------------------------------------------------------------------
#define DvcPwrInfoSelfPwr  0x01 //Device is Self-Powered
#define DvcPwrInfoHubPPP   0x40 //Device is Per-Port-Powered Hub
#define DvcPwrInfoDvcIsHub 0x80 //Device is a Hub


// USB transaction directions
#define USBDIR_HOSTTODEVICE 0
#define USBDIR_DEVICETOHOST 1


// USB control/setup request types
#define USBCTREQTYPE_STANDARD 0x00
#define USBCTREQTYPE_CLASS    0x20
#define USBCTREQTYPE_VENDOR   0x40
 

// USB notification codes
#define USBNOTIFICATION_INTERFACECHANGE 1
#define USBNOTIFICATION_CONTROLMESSAGE 2
#define USBNOTIFICATION_BULKMESSAGE 3
#define USBNOTIFICATION_PERIODICINTERRUPT 4


/////////////////////////////////////
// Structures and global variables //
/////////////////////////////////////


typedef struct usbrequest_s
{
   // int 14h USB request structure

   unsigned char I14RRequestType;     //Request Type
   unsigned char I14RFlags;           //Bit-level Flags
   unsigned char I14RHostIndex;       //Host Index
   unsigned char I14RDeviceAddress;   //USB Device Address
                                      //  Also I14RHubAddress
   unsigned char I14REndPoint;        //End Point
                                      //  Also I14RHubPort
                                      //  Also I14RAltInterface
   unsigned char I14RConfigValue;     //Configuration Value
                                      //  Also I14RCloseID
   unsigned char I14RInterfaceNum;    //Interface Number
   unsigned char I14RSearchIndex;     //Search Index
   unsigned int  I14RVendorID;        //Vendor ID
   unsigned int  I14RProductID;       //Product ID
   unsigned char I14RDvcClass;        //Device Class
   unsigned char I14RDvcSubClass;     //Device SubClass
   unsigned char I14RDvcProtocol;     //Device Protocol
   unsigned char I14RIntfClass;       //Interface Class
   unsigned char I14RIntfSubClass;    //Interface SubClass
   unsigned char I14RIntfProtocol;    //Interface Protocol
   unsigned int  I14RRequestHandle;   //Request Handle Number
   unsigned int  I14RPeriodicity;     //Interrupt Periodicity/Duration
                                      //  Also I14RBeepFrequency
   unsigned int  I14RTimeout;         //Transaction Timeout Value
   unsigned int  I14RDataAddrOff;     //Data Address Offset
   unsigned int  I14RDataAddrSeg;     //Data Address Segment
   unsigned int  I14RDataSize;        //Size of Data (bytes)
   unsigned int  I14RCallBackAddrOff; //Call Back Address Offset (IP)
                                      //  Also I14RLargeCallRtnCode
   unsigned int  I14RCallBackAddrSeg; //Call Back Address Segment (CS)
   unsigned int  I14RUserPktID;       //User Packet ID
   unsigned char I14RSetupReqType;    //Setup Packet Request Type
   unsigned char I14RSetupReq;        //Setup Packet Request Code
   unsigned int  I14RSetupValue;      //Setup Packet Value
   unsigned int  I14RSetupIndex;      //Setup Packet Index
   unsigned int  I14RSetupLength;     //Setup Packet Length
   unsigned int  I14RFrameTiming;     //Frame Timing Value (def = 12000)
                                      //  Also I14RFrameIndex
   unsigned int  I14RIsochSchedOff;   //Isochronous Schedule Table Offset
   unsigned int  I14RIsochSchedSeg;   //Isochronous Schedule Table Segment
   unsigned int  I14RFiller1;         //Fillers to equal 64 bytes
   unsigned int  I14RFiller2;         //  "     "    "   "    "
   unsigned int  I14RFiller3;         //  "     "    "   "    "
   unsigned int  I14RFiller4;         //  "     "    "   "    "
   unsigned int  I14RFiller5;         //  "     "    "   "    "
   unsigned int  I14RFiller6;         //  "     "    "   "    "
   unsigned int  I14RFiller7;         //  "     "    "   "    "
} usbrequest_t;


typedef struct usbreply_s
{
   unsigned int ax; // processor AX register after the driver call returns
   unsigned int bx; // processor BX register after the driver call returns
   unsigned int cx; // processor CX register after the driver call returns
   unsigned int dx; // processor DX register after the driver call returns
} usbreply_t;


typedef struct usbnotification_s
{
   int status;
   int id;
   int reg_cx;
   int reg_dx;
} usbnotification_t;


typedef struct usbhostinfo_s
{
   unsigned int vendor_id;
   unsigned int product_id;
   char host_type; // 'U', 'O', 'E' or 'X'HCI, or '?' (unknown)
   char is_pcibus;
   int irq_number;
   int roothub_ports;
} usbhostinfo_t;


typedef struct usbdeviceinfo_s
{
   unsigned int vendor_id;
   unsigned int product_id;
   int device_class;
   int device_subclass;
   int device_protocol;
   int usbversion_major;
   int usbversion_minor;
   int hub_address; // 0xFF if root hub
   int hub_port;
   int config_value;
   int power_draw;
   char is_selfpowered;
   char is_testmode_enabled;
   char is_remotewakeup_enabled;
   char is_remotewakeup_supported;
   char is_lowspeed;
   char being_configured; // NOTE: if being configured, only hub_address, hub_port & is_lowspeed are valid
   char is_bad;
   int baddevice_flag;
   int baddevice_err;
} usbdeviceinfo_t;


/////////////////////////////////////////////////
// Global variables used by the driver wrapper //
/////////////////////////////////////////////////


#ifdef USB_GLOBALS
static usbrequest_t usb_request, *usb_request_ptr = &usb_request; // USB request singleton (needs to be global)
static usbreply_t usb_reply; // USB reply singleton (needs to be global)
static usbnotification_t usb_notifications[32]; // USB notifications array singleton (needs to be global)
static int usb_notification_count = 0;
#endif


///////////////
// Functions //
///////////////


// prototypes of featured functions
#ifndef USB_GLOBALS
extern int USB_IsDriverPresent (void);
extern int USB_GetHostInfo (int host_index, usbhostinfo_t *host_info);
extern int USB_GetDeviceInfo (int host_index, int device_address, usbdeviceinfo_t *device_info);
extern int USB_RegisterAsOwner (int host_index, int device_address, int device_interface, unsigned int *out_handle);
extern int USB_UnregisterAsOwner (int host_index, unsigned int owner_handle);
extern int USB_ScheduleInterrupt (int host_index, int device_address, int device_endpoint, int periodicity, int frame_index, 
                                  int direction, char *data, int datalen, unsigned int *out_txnhandle);
extern int USB_SendControl (int host_index, int device_address, int device_endpoint, int timeout,
                            unsigned char setup_reqtype, unsigned char setup_req, unsigned short setup_value, unsigned short setup_index,
                            char *data, int datalen, unsigned int *out_txnhandle, unsigned char *out_closureid);
//extern int USB_SendBulk (int host_index, int device_address, int device_endpoint, int timeout,
//                         char *data, int datalen);
extern int USB_CancelTransaction (int host_index, unsigned int txn_handle, unsigned char closure_id);
extern int USB_PopUnprocessedNotification (usbnotification_t *out_notification);
#endif // !USB_GLOBALS


#ifdef USB_GLOBALS
static int USB_DoRequest (void)
{
   // this function sends a "standard" Int14h request to the USB driver

   static union REGS regs;

   regs.x.ax = 0x5001; // AX = requested function (Do Function)
   regs.x.bx = 0x5553; // BX = "US" (magic number for the USB driver hook)
   regs.x.cx = 0x4221; // CX = "B!" (magic number for the USB driver hook)
   regs.x.dx = FP_OFF (usb_request_ptr); // DS:[DX] = request structure

   int86 (0x14, &regs, &regs); // call Int14h (USB driver hooks it)
#ifdef USB_DEBUG
   if (regs.x.ax != 0)
   {
      fprintf (stderr, "USB_DoRequest(0x%04x) error!\n", usb_request.I14RRequestType);
      fprintf (stderr, "  AX: 0x%04x (%d)\n", regs.x.ax, regs.x.ax);
      fprintf (stderr, "  BX: 0x%04x (%d)\n", regs.x.bx, regs.x.bx);
      fprintf (stderr, "  CX: 0x%04x (%d)\n", regs.x.cx, regs.x.cx);
      fprintf (stderr, "  DX: 0x%04x (%d)\n", regs.x.dx, regs.x.dx);
   }
#endif

   usb_reply.ax = regs.x.ax; // set return AX
   usb_reply.bx = regs.x.bx; // set return BX
   usb_reply.cx = regs.x.cx; // set return CX
   usb_reply.dx = regs.x.dx; // set return DX

   return (regs.x.ax); // and pass back called function's return value
}


static void far USB_NotificationCallback (void)
{
   // this is a "generic" USB notification catcher. It stacks up to a maximum amounts of notifications
   // received from the USB subsystem until the main loop of your program is ready to process them.
   // The AX (status), BX (user-supplied argument to the callback, saved as 'id'), CX and DX registers
   // are saved in a usbnotification_t structure.

   int status;
   int id;
   int reg_cx;
   int reg_dx;

   _asm { mov status, ax } // retrieve the status
   _asm { mov id, bx } // retrieve the notification ID
   _asm { mov reg_cx, cx } // retrieve the other notification registers
   _asm { mov reg_dx, dx } // retrieve the other notification registers

   if ((id == USBNOTIFICATION_PERIODICINTERRUPT) && (status == TDStsNAKReceived))
      return; // don't stack up periodic interrupt notifications for which there is no data

   if (usb_notification_count >= sizeof (usb_notifications) / sizeof (&usb_notifications[0]))
      return; // no space left for notification !

   // stack up this notification in the global USB notifications array
   usb_notifications[usb_notification_count].status = status;
   usb_notifications[usb_notification_count].id = id;
   usb_notifications[usb_notification_count].reg_cx = reg_cx;
   usb_notifications[usb_notification_count].reg_dx = reg_dx;
   usb_notification_count++;

   return; // finished
}


int USB_IsDriverPresent (void)
{
   // this function tests to see if a compatible USB host controller driver
   // is loaded in memory. Returns TRUE if so, FALSE otherwise.

   static union REGS regs;

   regs.x.ax = 0x5000; // AX = requested function (Installation Check)
   regs.x.bx = 0x5553; // BX = "US" (magic number for the USB driver hook)
   regs.x.cx = 0x4221; // CX = "B!" (magic number for the USB driver hook)
   int86 (0x14, &regs, &regs); // call Int 14h (USB driver hooks it)

   // and return whether the BX and CX registers are now swapped as expected
   return ((regs.x.ax == 0x0000) && (regs.x.bx == 0x4221) && (regs.x.cx == 0x5553));
}


int USB_GetHostInfo (int host_index, usbhostinfo_t *host_info)
{
   // this function dumps info for an individual USB host.
   // Returns TRUE on success (host was valid), FALSE if host was invalid.

   // issue GetHostVendInfo request
   usb_request.I14RRequestType = I14RRTGetHostVendInfo;
   usb_request.I14RHostIndex = host_index;
   if (USB_DoRequest () != 0)
      return (0); // return FALSE if the USB command was refused

   // USB request was successful, save results
   host_info->vendor_id = usb_reply.bx;
   host_info->product_id = usb_reply.cx;

   // issue GetHostHWInfo request
   usb_request.I14RRequestType = I14RRTGetHostHWInfo;
   usb_request.I14RHostIndex = host_index;
   if (USB_DoRequest () != 0)
      return (0); // return FALSE if the USB command was refused

   // USB request was successful, save results
   if      ((usb_reply.bx & 0xF) == USBHostTypeUHCI) host_info->host_type = 'U';
   else if ((usb_reply.bx & 0xF) == USBHostTypeOHCI) host_info->host_type = 'O';
   else if ((usb_reply.bx & 0xF) == USBHostTypeEHCI) host_info->host_type = 'E';
   //else if ((usb_reply.bx & 0xF) == USBHostTypeXHCI) host_info->host_type = 'X';
   else                                              host_info->host_type = '?';
   host_info->is_pcibus = ((usb_reply.bx & USBBusTypePCI) != 0);
   host_info->irq_number    = ((usb_reply.cx >> 0) & 0xFF);
   host_info->roothub_ports = ((usb_reply.cx >> 8) & 0xFF);

   return (1); // USB command was accepted, return TRUE
}


int USB_GetDeviceInfo (int host_index, int device_address, usbdeviceinfo_t *device_info)
{
   // this function gathers the info for a single USB device (at host/address)
   // Returns TRUE on success, FALSE otherwise

   int InterfaceNum;        //Interface Number
   int AltInterface;        //Alternate Interface Number
   int NumSpaces;

   // issue GetDvcVendInfo request
   usb_request.I14RRequestType = I14RRTGetDvcVendInfo;
   usb_request.I14RHostIndex = host_index;
   usb_request.I14RDeviceAddress = device_address;
   if (USB_DoRequest () != 0)
      return (0); // return FALSE if the USB command was refused

   // USB request was successful, save results
   device_info->vendor_id = usb_reply.bx;
   device_info->product_id = usb_reply.cx;

   // issue GetDvcClassInfo request
   usb_request.I14RRequestType = I14RRTGetDvcClassInfo;
   usb_request.I14RHostIndex = host_index;
   usb_request.I14RDeviceAddress = device_address;
   if (USB_DoRequest () != 0)
      return (0); // return FALSE if the USB command was refused

   // USB request was successful, save results
   device_info->device_class    = ((usb_reply.bx >> 0) & 0xFF);
   device_info->device_subclass = ((usb_reply.bx >> 8) & 0xFF);
   device_info->device_protocol  = ((usb_reply.cx >>  0) & 0xFF);
   device_info->usbversion_minor = ((usb_reply.cx >>  8) & 0x0F);
   device_info->usbversion_major = ((usb_reply.cx >> 12) & 0x0F);
   device_info->is_lowspeed = ((usb_reply.dx & LowSpeedFlag) != 0);

   // issue GetDvcHubInfo request
   usb_request.I14RRequestType = I14RRTGetDvcHubInfo;
   usb_request.I14RHostIndex = host_index;
   usb_request.I14RDeviceAddress = device_address;
   if (USB_DoRequest () != 0)
      return (0); // return FALSE if the USB command was refused

   // USB request was successful, save results
   device_info->hub_address = ((usb_reply.bx >> 0) & 0xFF);
   device_info->hub_port    = ((usb_reply.bx >> 8) & 0xFF);

   // issue GetDvcPowerInfo request
   usb_request.I14RRequestType = I14RRTGetDvcPowerInfo;
   usb_request.I14RHostIndex = host_index;
   usb_request.I14RDeviceAddress = device_address;
   if (USB_DoRequest () != 0)
      return (0); // return FALSE if the USB command was refused

   // USB request was successful, save results
   device_info->is_selfpowered = ((usb_reply.cx & DvcPwrInfoSelfPwr) != 0);
   device_info->power_draw = usb_reply.bx;

   // issue GetDvcStatus request
   usb_request.I14RRequestType = I14RRTGetDvcStatus;
   usb_request.I14RHostIndex = host_index;
   usb_request.I14RDeviceAddress = device_address;
   if (USB_DoRequest () != 0)
      return (0); // return FALSE if the USB command was refused

   // USB request was successful, save results
   device_info->being_configured          = (((usb_reply.bx >> 8) & BeingConfiged)    != 0);
   device_info->is_bad                    = (((usb_reply.bx >> 8) & BadDevice)        != 0);
   device_info->is_testmode_enabled       = (((usb_reply.bx >> 8) & TestModeEnabled)  != 0);
   device_info->is_remotewakeup_enabled   = (((usb_reply.bx >> 8) & RWakeupEnabled)   != 0);
   device_info->is_remotewakeup_supported = (((usb_reply.bx >> 8) & RWakeupSupported) != 0);
   device_info->config_value = (usb_reply.bx & 0xFF);
   device_info->baddevice_flag = (device_info->is_bad ? (usb_reply.cx & 0xFF) : 0);
   device_info->baddevice_err  = (device_info->is_bad ? usb_reply.dx : 0);

   return (1); // USB command was accepted, return TRUE
}


int USB_RegisterAsOwner (int host_index, int device_address, int device_interface, unsigned int *out_handle)
{
   // prepare request structure
   usb_request.I14RRequestType = I14RRTRegIntfOwner;
   usb_request.I14RHostIndex = host_index;
   usb_request.I14RDeviceAddress = device_address;
   usb_request.I14RInterfaceNum = device_interface;
   usb_request.I14RCallBackAddrSeg = FP_SEG (USB_NotificationCallback); // segment
   usb_request.I14RCallBackAddrOff = FP_OFF (USB_NotificationCallback); // offset
   usb_request.I14RUserPktID = USBNOTIFICATION_INTERFACECHANGE;
   if (USB_DoRequest () != 0)
      return (0); // return FALSE if the USB command was refused

   *out_handle = usb_reply.cx; // return our new owner handle
   return (1); // USB command was accepted, return TRUE
}


int USB_UnregisterAsOwner (int host_index, unsigned int owner_handle)
{
   // prepare request structure
   usb_request.I14RRequestType = I14RRTUnRegIntfOwner;
   usb_request.I14RHostIndex = host_index;
   usb_request.I14RRequestHandle = owner_handle;
   if (USB_DoRequest () != 0)
      return (0); // return FALSE if the USB command was refused

   return (1); // USB command was accepted, return TRUE
}


int USB_ScheduleInterrupt (int host_index, int device_address, int device_endpoint, int periodicity, int frame_index, 
                           int direction, char *data, int datalen, unsigned int *out_txnhandle)
{
   // prepare request structure
   usb_request.I14RRequestType = I14RRTDoInterruptPer; // request Type = 0x95 Schedule Periodic Interrupt
   usb_request.I14RHostIndex = host_index;
   usb_request.I14RDeviceAddress = device_address;
   usb_request.I14REndPoint = device_endpoint;
   usb_request.I14RPeriodicity = periodicity;
   usb_request.I14RDataAddrSeg = FP_SEG (data); // segment
   usb_request.I14RDataAddrOff = FP_OFF (data); // offset
   usb_request.I14RDataSize = datalen;
   usb_request.I14RFrameTiming = frame_index;
   usb_request.I14RFlags = (direction == USBDIR_DEVICETOHOST ? 0x01 : 0x00) | (frame_index != -1 ? 0x40 : 0x00);
   usb_request.I14RCallBackAddrSeg = FP_SEG (USB_NotificationCallback); // segment
   usb_request.I14RCallBackAddrOff = FP_OFF (USB_NotificationCallback); // offset
   usb_request.I14RUserPktID = USBNOTIFICATION_PERIODICINTERRUPT;

   if (USB_DoRequest () != 0)
      return (0); // return FALSE if the USB command was refused

   if (out_txnhandle != NULL) *out_txnhandle = usb_reply.cx; // return our new transaction handle
   return (1); // USB command was accepted, return TRUE
}


int USB_SendControl (int host_index, int device_address, int device_endpoint, int timeout,
                     unsigned char setup_reqtype, unsigned char setup_req, unsigned short setup_value, unsigned short setup_index,
                     char *data, int datalen, unsigned int *out_txnhandle, unsigned char *out_closureid)
{
   // prepare request structure
   usb_request.I14RRequestType = I14RRTDoControl; // request Type = 0x96 Schedule Control/Setup Request
   usb_request.I14RHostIndex = host_index;
   usb_request.I14RDeviceAddress = device_address;
   usb_request.I14REndPoint = device_endpoint;
   usb_request.I14RTimeout = 10 * timeout; // may be 0, in which case a 5s timeout will be chosen
   usb_request.I14RFlags = 0; // data direction: Host to Device
   usb_request.I14RSetupReqType = setup_reqtype; // Setup Packet Request Type
   usb_request.I14RSetupReq = setup_req; // Setup Packet Request Code
   usb_request.I14RSetupValue = setup_value; // Setup Packet Value
   usb_request.I14RSetupIndex = setup_index; // Setup Packet Index
   usb_request.I14RDataAddrSeg = FP_SEG (data); // segment
   usb_request.I14RDataAddrOff = FP_OFF (data); // offset
   usb_request.I14RSetupLength = datalen; // Setup Packet Length
   usb_request.I14RCallBackAddrSeg = FP_SEG (USB_NotificationCallback); // segment
   usb_request.I14RCallBackAddrOff = FP_OFF (USB_NotificationCallback); // offset
   usb_request.I14RUserPktID = USBNOTIFICATION_CONTROLMESSAGE;

   if (USB_DoRequest () != 0)
      return (0); // return FALSE if the USB command was refused

   if (out_txnhandle != NULL) *out_txnhandle = usb_reply.cx; // return our new transaction handle
   if (out_closureid != NULL) *out_closureid = (unsigned char) (usb_reply.dx & 0x00FF); // and our new closure ID
   return (1); // USB command was accepted, return TRUE
}

/*
int USB_SendBulk (int host_index, int device_address, int device_endpoint, int timeout,
                  char *data, int datalen,
                  void *callback_address, unsigned int callback_arg)
{
   // prepare request structure
   usb_request.I14RRequestType = I14RRTDoBulk; // request Type = 0x97 Schedule Bulk Transaction
   usb_request.I14RHostIndex = host_index;
   usb_request.I14RDeviceAddress = device_address;
   usb_request.I14REndPoint = device_endpoint;
   usb_request.I14RTimeout = 10 * timeout; // may be 0, in which case a 5s timeout will be chosen
   usb_request.I14RFlags = 0; // data direction: Host to Device
   usb_request.I14RDataAddrSeg = FP_SEG (data); // segment
   usb_request.I14RDataAddrOff = FP_OFF (data); // offset
   usb_request.I14RDataSize = datalen;
   usb_request.I14RCallBackAddrSeg = FP_SEG (USB_NotificationCallback); // segment
   usb_request.I14RCallBackAddrOff = FP_OFF (USB_NotificationCallback); // offset
   usb_request.I14RUserPktID = USBNOTIFICATION_BULKMESSAGE;
   if (USB_DoRequest () != 0)
      return (0); // return FALSE if the USB command was refused

   return (1); // USB command was accepted, return TRUE
}


int USB_RecvBulk (int host_index, int device_index, int device_endpoint, int timeout,
                  char *data, int datalen,
                  void *callback_address, unsigned int callback_arg)
{
   // FIXME: I fail to see how this can work.

   // prepare request structure
   usb_request.I14RRequestType = I14RRTDoBulk; // request Type = 0x97 Schedule Bulk Transaction
   usb_request.I14RHostIndex = host_index;
   usb_request.I14RDeviceAddress = device_index;
   usb_request.I14REndPoint = device_endpoint;
   usb_request.I14RTimeout = 10 * timeout; // may be 0, in which case a 5s timeout will be chosen
   usb_request.I14RFlags = I14RFlagIn; // data direction: device to host
   usb_request.I14RDataAddrSeg = FP_SEG (data); // segment
   usb_request.I14RDataAddrOff = FP_OFF (data); // offset
   usb_request.I14RDataSize = datalen;
   usb_request.I14RCallBackAddrSeg = FP_SEG (USB_NotificationCallback); // segment
   usb_request.I14RCallBackAddrOff = FP_OFF (USB_NotificationCallback); // offset
   usb_request.I14RUserPktID = USBNOTIFICATION_BULKMESSAGE;
   if (USB_DoRequest () != 0)
      return (0); // return FALSE if the USB command was refused

   return (1); // USB command was accepted, return TRUE
}
*/

int USB_CancelTransaction (int host_index, unsigned int txn_handle, unsigned char closure_id)
{
   // prepare request structure
   usb_request.I14RRequestType = I14RRTCloseHandle; // request Type = 0x9C Close/Remove Scheduled Transaction
   usb_request.I14RHostIndex = host_index;
   usb_request.I14RRequestHandle = txn_handle;
   usb_request.I14RConfigValue = closure_id;
   if (USB_DoRequest () != 0)
      return (0); // return FALSE if the USB command was refused

   return (1); // USB command was accepted, return TRUE
}


int USB_PopUnprocessedNotification (usbnotification_t *out_notification)
{
   // this function pops one USB notification from the unprocessed USB notifications stack and copies it out
   // in out_notification, decreasing the global count of unprocessed notifications (i.e. freeing the array slot).

   if (usb_notification_count == 0)
      return (0); // return FALSE if the notifications array contains no notification

   memcpy (out_notification, &usb_notifications[usb_notification_count - 1], sizeof (usbnotification_t));
   usb_notification_count--; // one notification less in this array

   return (1); // we could successfully return an unprocessed notification, return TRUE
}


#endif // USB_GLOBALS


#endif // USB_H
