//--------------------------------------------------------------------------
// eGalax USB touch panel driver program (this program is NOT a TSR).
// This driver requires USBUHCI to be loaded. This program will also NOT work
//   with the USB in Legacy Mode (i.e, when USB is controlled by the BIOS).
//--------------------------------------------------------------------------


// compile with Borland Turbo-C++ 3.0
// MANDATORY COMPILER OPTIONS: -a -mt
//                              :  :
//                              :  +.. tiny memory model
//                              +........ word alignment


// standard C includes
#include <stdio.h>
#include <time.h>
#include <dos.h>
#ifdef _MSC_VER
#define kbhit _kbhit
#define _asm __asm
#endif


// our own includes
#define USB_GLOBALS
#define USB_DEBUG
#include "usb.h"


// miscellaneous definitions
#define OUR_PRODUCT_NAME "eGalax USB Touch Panel" // FIXME: we should read this from its descriptor tables


// boolean type support
#ifndef bool
#define bool int
#define false 0
#define true 1
#endif


// function prototypes
void ProcessDiagnosticPacket (char *packet_data, int packet_length);
void ProcessReportPacket (char *packet_data, int packet_length);
bool QueryFirmwareVersion (int host_index, int device_address);
bool QueryFirmwareType (int host_index, int device_address);
void millisleep (int millisec);

// global variables
// -------- CONFIGURATION BEGIN --------
int our_vendor_id = 0x0EEF; // the touch panel's vendor ID
int our_product_id = 0x0001; // the touch panel's product ID
bool want_axis_swap = true; // set to 1 if X and Y axes should be swapped
bool want_axis_xflip = true; // set to 1 if X axis should be flipped (left becomes right and vice versa)
bool want_axis_yflip = false; // set to 1 if Y axis should be flipped (top becomes bottom and vice versa)
// -------- CONFIGURATION END --------
char status_packet[8]; // the device descriptor table tells us it can report 8 bytes max in a packet
char firmware_version[8]; // a buffer to hold the firmware version, if the device decides to send it
char firmware_type[8]; // a buffer to hold the firmware type, if the device decides to send it
bool is_touchdown; // TRUE if the surface is currently touched, FALSE otherwise
int touch_x; // touch position fraction of the surface, ranging from 0 (left) to 2^RESOLUTION_BITS (right)
int touch_y; // touch position fraction of the surface, ranging from 0 (top) to 2^RESOLUTION_BITS (bottom)
int pressure_value; // vendor-specific info.
int player_id; // vendor-specific info. No idea what it means.


int main (int argc, char **argv)
{
   // program entrypoint

   usbhostinfo_t host_info;
   usbdeviceinfo_t device_info;
   usbnotification_t usb_notification;
   int host_index; // host controller index (0-15)
   int device_address; // USB device address (1-127)
   unsigned int owner_handle; // interface owner handle, given to us by USBUHCI
   unsigned int txn_handle; // transaction handle, given to us by USBUHCI
   int packet_length;
   int byte_index;

   // silence Borland compiler warning
   (void) argc;
   (void) argv;

   // first, ensure that the USB driver is present
   if (!USB_IsDriverPresent ())
   {
      fprintf (stderr, "No compatible USB driver is loaded. Please load USBUHCI.\n");
      exit (1); // bomb out if USBUHCI driver is not loaded
   }

   // try and locate our USB device, cycle through all possible hosts...
   for (host_index = 0; host_index < 16; host_index++)
   {
      if (!USB_GetHostInfo (host_index, &host_info))
         continue; // skip invalid/unexistent hosts

      // cycle through all possible device addresses on this host
      for (device_address = 1; device_address < 128; device_address++)
      {
         if (!USB_GetDeviceInfo (host_index, device_address, &device_info))
            continue; // skip invalid/unexistent device addresses

         if ((device_info.vendor_id == our_vendor_id) && (device_info.product_id == our_product_id))
            break; // if this device is our USB device, stop searching
      }

      if (device_address != 128)
         break; // if the USB device was found, stop looping through hosts
   }

   // have we NOT found our device ?
   if (host_index == 16)
   {
      fprintf (stderr, "The " OUR_PRODUCT_NAME " was not detected.\n");
      exit (1); // bomb out if the USB device we want to drive is not attached
   }

   // have we found it on anything other than a UHCI host ?
   if (host_info.host_type != 'U')
   {
      fprintf (stderr, "The " OUR_PRODUCT_NAME " was detected on a %cHCI USB host chip.\n"
                       "This driver operates on UHCI hosts only. Please plug the device into a UHCI host.",
                        host_info.host_type);
      exit (1); // bomb out if the USB device we want to drive is attached to an incompatible host
   }

   // DEBUG
   printf ("Touch panel found on host %d addr %d\n", host_index, device_address);

   // register ourselves as the owner of this device
   if (!USB_RegisterAsOwner (host_index, device_address, 0, &owner_handle))
   {
      fprintf (stderr, "Fatal error: " OUR_PRODUCT_NAME " ownership"
                       " registration was refused to us by USBUHCI.\n");
      exit (1); // bomb out if the ownership can't be granted to us
   }

   // DEBUG
   printf ("Registration successful, owner handle = 0x%04x\n", owner_handle);

   // TODO: READ INPUT ENDPOINT FROM DESCRIPTOR INSTEAD OF RELYING ON HARDCODED VALUES

   //   END POINT DESCRIPTOR
   // =========================
   // Descr Length:        0x07  7
   // Descr Type:          0x05  EndPoint
   // End Point Number:    0x81  EndPoint = 1  Direction = In
   // Attributes:          0x03  Interrupt NonSynchronous Data
   // Max Packet Size:   0x0008  8
   // Max Poll Interval:   0x05  5

   // the endpoint descriptor table for this USB device tells us that it uses the interrupt-based periodic transaction
   // mode to communicate with us. So schedule a periodic poll of its USB endpoint, that will call our callback function
   // at a reasonable framerate.
   if (!USB_ScheduleInterrupt (host_index, device_address, 1 /*endpoint*/, 5 /*periodicity*/, -1 /*frame index*/,
                               USBDIR_DEVICETOHOST, status_packet, 8 /*expected max data size*/, &txn_handle))
   {
      fprintf (stderr, "Fatal error: couldn't schedule periodic interrupt setup packet to device.\n");
      USB_UnregisterAsOwner (host_index, owner_handle);
      exit (1); // if we can't do that, bomb out
   }

   // the second thing to do is to send a control/setup packet to the device to instruct it that we want it to talk
   // over the USB bus and not over the RS232 serial one. This is done by sending it a vendor-specific init string.
   if (!USB_SendControl (host_index, device_address, 0 /*endpoint*/, 0 /*timeout*/,
                         /*setup packet arguments:*/ USBCTREQTYPE_VENDOR /*reqtype*/, 0 /*req*/, 0 /*value*/, 0 /*index*/,
                         "\n1A" /*packet data*/, 3 /*packet length*/, NULL /*(out)txn handle*/, NULL /*(out)closure ID*/))
   {
      fprintf (stderr, "Fatal error: couldn't send initialization packet to device.\n");
      USB_CancelTransaction (host_index, txn_handle, 0);
      USB_UnregisterAsOwner (host_index, owner_handle);
      exit (1); // if we can't do that, bomb out
   }

   // DEBUG
   printf (OUR_PRODUCT_NAME " driver running. Press a key to abort.\n");

   // main program loop
   while (!kbhit ())
   {
      while (USB_PopUnprocessedNotification (&usb_notification))
      {
         if (usb_notification.id == USBNOTIFICATION_INTERFACECHANGE)
         {
            printf ("==> Received interface notification!\n");
            printf ("    status = %d    id = %d    reg_cx = 0x%04x    reg_dx = 0x%04x\n",
                    usb_notification.status, usb_notification.id, usb_notification.reg_cx, usb_notification.reg_dx);
         }
         else if (usb_notification.id == USBNOTIFICATION_PERIODICINTERRUPT)
         {
            packet_length = usb_notification.reg_cx; // save packet length

            //printf ("==> Received data from device! %d bytes, status %d\n", packet_length, usb_notification.status);
            //printf ("    ");
            //for (byte_index = 0; byte_index < packet_length; byte_index++)
            //   printf ("%02x ", (unsigned char) status_packet[byte_index]);
            //printf ("        ");
            //for (byte_index = 0; byte_index < packet_length; byte_index++)
            //   printf ("%c", (status_packet[byte_index] > 31 ? (unsigned char) status_packet[byte_index] : 254));
            //printf ("\n");
            
            // is it a diagnostic packet or a report packet ?
            if (status_packet[0] == '\n')
               ProcessDiagnosticPacket (status_packet, packet_length);
            else if (status_packet[0] & 0x80)
               ProcessReportPacket (status_packet, packet_length);
         }
      }
      millisleep (100);
   }

   // if we reach here, it means the user pressed a key to exit the program

   // stop periodic transactions with this device
   if (!USB_CancelTransaction (host_index, txn_handle, 0))
   {
      fprintf (stderr, "Warning: " OUR_PRODUCT_NAME " periodic transactions"
                       " cancelling at USBUHCI failed.\n");
      exit (1); // bomb out if the ownership can't be released
   }

   // UNregister ourselves as the owner of the device
   if (!USB_UnregisterAsOwner (host_index, owner_handle))
   {
      fprintf (stderr, "Warning: " OUR_PRODUCT_NAME " ownership"
                       " unregistration at USBUHCI failed.\n");
      exit (1); // bomb out if the ownership can't be released
   }

   return (0); // finished, return to DOS
}


void ProcessDiagnosticPacket (char *packet_data, int packet_length)
{
   // this function processes what the device vendor calls a 'diagnostic' packet. Such packets
   // begin with '\n' and report miscellaneous and relatively useless data about the device.

   if (packet_length < 3)
   {
      //fprintf (stderr, "Warning: received invalid report packet ! (length: %d bytes)\n", packet_length);
      return; // consistency check
   }

   if (packet_data[2] == 'A')
      ; // this packet simply means the device is working properly
   else if (packet_data[2] == 'D')
   {
      strncpy (firmware_version, &packet_data[3], packet_data[1] - 1); // copy out the firmware version
      firmware_version[packet_data[1] - 1] = 0; // null-terminate the string ourselves
      
      printf ("Received firmware version: %s\n", firmware_version);
   }
   else if (packet_data[2] == 'E')
   {
      strncpy (firmware_type, &packet_data[3], packet_data[1] - 1); // copy out the firmware type
      firmware_type[packet_data[1] - 1] = 0; // null-terminate the string ourselves
      
      printf ("Received firmware type: %s\n", firmware_type);
   }
   //else
   //   fprintf (stderr, "Warning: received unknown diagnostic packet !\n");

   return; // we know no other sort of diagnostic packet
}


void ProcessReportPacket (char *packet_data, int packet_length)
{
   // this function processes what the device vendor calls a 'report' packet. Such packets
   // begin with the first byte's most significant bit set, and carry touch coordinates.

   bool has_pressureinfo;
   bool has_playerid;
   int resolution_bits;
   bool new_touchdown;
   int new_x;
   int new_y;

   if (packet_length < 5)
   {
      //fprintf (stderr, "Warning: received invalid report packet ! (length: %d bytes)\n", packet_length);
      return; // consistency check
   }

   has_pressureinfo = ((packet_data[0] & 0x40) != 0);
   has_playerid     = ((packet_data[0] & 0x20) != 0);
   resolution_bits = 11 + ((packet_data[0] >> 1) & 0x03);
   new_touchdown     = ((packet_data[0] & 0x01) != 0);

   new_x = ((packet_data[want_axis_swap ? 3 : 1] & 0x7F) << 7) | ((packet_data[want_axis_swap ? 4 : 2] & 0x7F) << 0);
   if (want_axis_xflip)
      new_x = (1 << resolution_bits) - new_x;

   new_y = ((packet_data[want_axis_swap ? 1 : 3] & 0x7F) << 7) | ((packet_data[want_axis_swap ? 2 : 4] & 0x7F) << 0);
   if (want_axis_yflip)
      new_y = (1 << resolution_bits) - new_y;

   if (has_pressureinfo && (packet_length >= 6))
      pressure_value = (packet_data[5] & 0x7F);
   else if (has_playerid && (packet_length >= 6))
      player_id = (packet_data[5] & 0x7F);

   // have the coordinates changed ? if so, update them
   if ((new_touchdown != is_touchdown) || (new_x != touch_x) || (new_y != touch_y))
   {
      is_touchdown = new_touchdown;
      touch_x = new_x;
      touch_y = new_y;

      printf ("Touch %s: %d, %d\n", (is_touchdown ? "DOWN" : "UP"), touch_x, touch_y);
   }

   return; // finished
}


bool QueryFirmwareVersion (int host_index, int device_address)
{
   // this function queries the device's firmware version, using a control/setup packet. It will reply through the
   // periodic IRQ-based transaction that we set up in the first place, and for which we already have a callback.

   return (USB_SendControl (host_index, device_address, 0 /*endpoint*/, 0 /*timeout*/,
                            /*setup packet arguments:*/ 0x40 /*reqtype*/, 0 /*req*/, 0 /*value*/, 0 /*index*/,
                            "\n1D" /*packet data*/, 3 /*packet length*/, NULL /*(out)txn handle*/, NULL /*(out)closure id*/));
}


bool QueryFirmwareType (int host_index, int device_address)
{
   // this function queries the device's firmware type, using a control/setup packet. It will reply through the
   // periodic IRQ-based transaction that we set up in the first place, and for which we already have a callback.

   return (USB_SendControl (host_index, device_address, 0 /*endpoint*/, 0 /*timeout*/,
                            /*setup packet arguments:*/ 0x40 /*reqtype*/, 0 /*req*/, 0 /*value*/, 0 /*index*/,
                            "\n1E" /*packet data*/, 3 /*packet length*/, NULL /*(out)txn handle*/, NULL /*(out)closure id*/));
}


void millisleep (int millisec)
{
   // handy function that sleeps a certain number of milliseconds, without overloading the processor

   clock_t start = clock ();

   while ((clock () - start) * 1000 / CLOCKS_PER_SEC < millisec)
      _asm {hlt}
}
